# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html
from pathlib import Path

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information

project = "sphinx-gha"
copyright = "2024, Andrew Cassidy"
author = "Andrew Cassidy"

# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

extensions = [
    "sphinx_gha",
    "myst_parser",
    "sphinx_examples",
    "sphinx_copybutton",
    "sphinx.ext.intersphinx",
]
default_role = "any"

templates_path = ["_templates"]
exclude_patterns = ["_build", "Thumbs.db", ".DS_Store"]
intersphinx_mapping = {"python": ("https://docs.python.org/3", None)}

# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

html_theme = "sphinx_rtd_theme"
html_static_path = []

# -- Options for sphinx-gha --------------------------------------------------

sphinx_gha_repo_root = str(Path(__file__).parent.parent.absolute())  # docs/..
sphinx_gha_repo_slug = "https://git.offworldcolonies.nexus/drewcassidy/sphinx-gha"

# -- Options for myst-parser -------------------------------------------------

myst_heading_anchors = 3
myst_enable_extensions = ["attrs_inline"]



from __future__ import annotations

import logging
import os
import typing as ty
from typing import Iterable

from docutils.nodes import Element
from sphinx import application
from sphinx.addnodes import pending_xref
from sphinx.builders import Builder
from sphinx.domains import Domain, ObjType
from sphinx.environment import BuildEnvironment
from sphinx.roles import XRefRole
from sphinx.util.nodes import make_refnode, find_pending_xref_condition

from sphinx_gha.git_ref import get_git_ref
from sphinx_gha.action import (
    ActionInputDirective,
    ActionOutputDirective,
    ActionEnvDirective,
    ActionDirective,
)
from sphinx_gha.common import ActionsExampleDirective
from sphinx_gha.workflow import (
    WorkflowInputDirective,
    WorkflowSecretDirective,
    WorkflowOutputDirective,
    WorkflowDirective,
)

logger = logging.getLogger(__name__)


class GHActionsDomain(Domain):
    name = "gha"
    label = "Github Actions"
    directives = {
        "action": ActionDirective,
        "action-input": ActionInputDirective,
        "action-output": ActionOutputDirective,
        "action-envvar": ActionEnvDirective,
        "workflow": WorkflowDirective,
        "workflow-input": WorkflowInputDirective,
        "workflow-secret": WorkflowSecretDirective,
        "workflow-output": WorkflowOutputDirective,
        "example": ActionsExampleDirective,
    }
    roles = {directive: XRefRole() for directive in directives}
    object_types = {role: ObjType(role) for role in roles.keys()}

    initial_data = {"objects": []}

    def get_full_qualified_name(self, node):
        parent_name = node.get("gh-actions:action") or node.get("gh-actions:workflow")
        target = node.get("reftarget")
        if target is None:
            return None
        else:
            return ".".join(filter(None, [parent_name, target]))

    def get_objects(self) -> Iterable[tuple[str, str, str, str, str, int]]:
        yield from self.data["objects"]

    def find_obj(
        self,
        env: BuildEnvironment,
        modname: str,
        classname: str,
        name: str,
        ty: str | None,
        searchmode: int = 0,
    ) -> list[tuple[str,]]:
        pass

    def resolve_xref(
        self,
        env: BuildEnvironment,
        fromdocname: str,
        builder: Builder,
        typ: str,
        target: str,
        node: pending_xref,
        contnode: Element,
    ) -> Element | None:
        typ_name = typ.split(":")[-1]
        dispname = target
        if len(target.split(".")) == 1:
            # figure out the parent object from the current context
            if typ_name.startswith("action-"):
                parent_name = node.get("gh-actions:action")
            elif typ_name.startswith("workflow-"):
                parent_name = node.get("gh-actions:workflow")
            else:
                parent_name = None
            target = ".".join(
                filter(None, [parent_name, target])
            )  # extend target full name with parent

        matches = [
            (docname, anchor)
            for name, dispname, objtyp, docname, anchor, prio in self.get_objects()
            if name == target and objtyp == typ
        ]

        if not matches:
            return None

        if len(matches) > 1:
            logger.warning(
                "more than one target found for cross-reference %r: %s",
                target,
                ", ".join(match[0] for match in matches),
            )

        docname, anchor = matches[0]

        # determine the content of the reference by conditions
        content = find_pending_xref_condition(node, "resolved")
        if content:
            children = content.children
        else:
            # if not found, use contnode
            children = [contnode]

        return make_refnode(
            builder, fromdocname, docname, anchor, children, title=dispname
        )

    def resolve_any_xref(
        self,
        env: BuildEnvironment,
        fromdocname: str,
        builder: Builder,
        target: str,
        node: pending_xref,
        contnode: Element,
    ) -> list[tuple[str, Element]]:
        matches = []
        for typ in self.object_types.keys():
            match = self.resolve_xref(
                env, fromdocname, builder, typ, target, node, contnode
            )
            if match is not None:
                matches.append((typ, match))
        return matches

    def note_object(self, name: str, dispname: str, typ: str, anchor: str) -> None:
        """Note a python object for cross-reference."""
        self.data["objects"].append((name, dispname, typ, self.env.docname, anchor, 1))


def setup(app: application.Sphinx) -> ty.Dict[str, ty.Any]:
    app.add_domain(GHActionsDomain)
    app.add_config_value("sphinx_gha_repo_root", os.getcwd(), "env")
    app.add_config_value(
        "sphinx_gha_repo_ref",
        lambda cfg: get_git_ref(cfg["sphinx_gha_repo_root"]),
        "env",
    )
    app.add_config_value("sphinx_gha_repo_slug", "UNKNOWN REPO", "env")
    app.setup_extension("myst_parser")

    return {
        "parallel_read_safe": True,
        "parallel_write_safe": True,
    }



def get_git_ref(repo_path):
    """
    Get the current git ref at the given repo path

    If this is a ReadTheDocs build environment, the version name will be used from there.
    Otherwise, it will use the current checked out tag or branch
    If none of these are true, returns the commit SHA
    """
    import git
    import os

    repo = git.Repo(repo_path)

    def get_branch():
        try:
            return repo.active_branch.name
        except TypeError:
            return None

    def get_tag():
        tags = [tag for tag in repo.tags if tag.commit == repo.head.commit]
        match tags:
            case []:
                return None
            case [tag, *_]:
                return tag.name

    def get_sha():
        return repo.head.commit.hexsha

    # see https://github.com/readthedocs/readthedocs.org/issues/11662,
    # this seems to be the best way to do this instead
    match os.environ.get("READTHEDOCS_VERSION_TYPE"):
        case None:
            return get_tag() or get_branch() or get_sha()
        case "branch":
            return os.environ.get("READTHEDOCS_GIT_IDENTIFIER")
        case "tag":
            return get_tag()
        case _:
            return get_sha()



from __future__ import annotations

import os
from functools import cached_property
from pathlib import Path

from docutils.parsers.rst import directives
from sphinx import addnodes
from sphinx.domains.std import StandardDomain
from sphinx.util.nodes import make_id

from sphinx_gha.common import (
    ActionsItemDirective,
    ActionsFileDirective,
    format_uses,
    yaml,
)


class ActionInputDirective(ActionsItemDirective):
    parent_role = "gh-actions:action"
    fields = ["required", "default"]
    option_spec = {"deprecationMessage": directives.unchanged_required}
    prefix = "input"

    def transform_content(self, content_node) -> None:
        super().transform_content(content_node)
        if deprecation_message := self.options.get("deprecationMessage"):
            admonition, msgs = self.format_deprecationMessage(deprecation_message)
            content_node.insert(0, admonition)


class ActionOutputDirective(ActionsItemDirective):
    parent_role = "gh-actions:action"
    prefix = "output"


class ActionEnvDirective(ActionsItemDirective):
    parent_role = "gh-actions:action"
    fields = ["required"]
    prefix = "env"

    def add_target_and_index(self, name: str, sig: str, signode) -> None:
        super().add_target_and_index(name, sig, signode)
        objtype = "envvar"
        node_id = make_id(self.env, self.state.document, objtype, name)
        signode["ids"].append(node_id)

        std: StandardDomain = self.env.domains["std"]
        std.note_object(objtype, signode["fullname"], node_id, location=signode)


class ActionDirective(ActionsFileDirective):
    role = "gh-actions:action"
    file_type = "action"

    @cached_property
    def path(self):
        path = self.options.get("path")
        if path is None:
            return None
        repo_root = Path(
            self.env.config["sphinx_gha_repo_root"] or os.getcwd()
        ).absolute()
        path = repo_root / Path(path)
        for filename in ["action.yml", "action.yaml"]:
            test_path = path / filename
            if test_path.exists():
                return test_path
        if path.is_file():
            return path

        self.error(f"Could not find an action definition at {path}")

    @classmethod
    def id_from_path(cls, path: Path):
        return path.parent.name

    @cached_property
    def example(self):
        if example_yaml := self.yaml.get("x-example"):
            return example_yaml

        if self.path is None:
            return ""

        slug = format_uses(self.path.parent, self.env)
        name = self.yaml.get("x-example-name")
        inputs = self.yaml.get("x_example_inputs") or {}
        env = self.yaml.get("x_example_env") or {}

        for k, d, e in [("x-env", env, "example"), ("inputs", inputs, "x-example")]:
            if action_inputs := self.yaml.get(k):
                for input_name, input_meta in action_inputs.items():
                    input_meta = input_meta or {}
                    if input_example := input_meta.get(e):
                        d[input_name] = input_example

        example_yaml = {}

        if name:
            example_yaml["name"] = name
        example_yaml["uses"] = slug
        if inputs:
            example_yaml["with"] = inputs
        if env:
            example_yaml["env"] = env

        example_yaml = [example_yaml]
        return yaml.dump(example_yaml)

    def transform_content(self, content_node: addnodes.desc_content) -> None:
        super().transform_content(content_node)

        # Items

        item_lists = [
            ("inputs", "Inputs", "action-input"),
            ("outputs", "Outputs", "action-output"),
            ("x-env", "Environment Variables", "action-envvar"),
        ]

        for key, title, directive in item_lists:
            if item_list := self.yaml.get(key):
                content_node.append(self.format_item_list(title, directive, item_list))



from __future__ import annotations

from functools import cached_property

from sphinx import addnodes

from sphinx_gha.common import (
    ActionsItemDirective,
    ActionsFileDirective,
    format_uses,
    yaml,
)


class WorkflowInputDirective(ActionsItemDirective):
    parent_role = "gh-actions:workflow"
    fields = ["required", "default", "type"]
    prefix = "input"


class WorkflowSecretDirective(WorkflowInputDirective):
    prefix = "secret"


class WorkflowOutputDirective(ActionsItemDirective):
    parent_role = "gh-actions:workflow"
    prefix = "output"


class WorkflowDirective(ActionsFileDirective):
    role = "gh-actions:workflow"
    file_type = "workflow"

    @cached_property
    def call_node(self):
        if (on_node := self.yaml.get("on") or self.yaml.get(True)) is None:
            # fucking yaml parses `on` as a boolean even in keys what the fuck
            return self.error(f"Workflow {self.path} has no `on` node")
        if "workflow_call" not in on_node:
            return self.error(f"Workflow {self.path} is not callable")
        else:
            return on_node.get("workflow_call") or {}

    @cached_property
    def example(self):
        if example_yaml := self.yaml.get("x-example"):
            return example_yaml

        slug = format_uses(self.path, self.env)
        name = self.yaml.get("x-example-name")
        inputs = self.yaml.get("x_example_inputs") or {}
        secrets = self.yaml.get("x_example_secrets") or {}

        for k, d, e in [
            ("secrets", secrets, "x-example"),
            ("inputs", inputs, "x-example"),
        ]:
            if action_inputs := self.call_node.get(k):
                for input_name, input_meta in action_inputs.items():
                    input_meta = input_meta or {}
                    if input_example := input_meta.get(e):
                        d[input_name] = input_example

        example_yaml = {}

        if name:
            example_yaml["name"] = name
        example_yaml["uses"] = slug
        if inputs:
            example_yaml["with"] = inputs
        if secrets:
            example_yaml["secrets"] = secrets

        example_yaml = [example_yaml]
        return yaml.dump(example_yaml)

    def transform_content(self, content_node: addnodes.desc_content) -> None:
        super().transform_content(content_node)
        # Items

        item_lists = [
            ("inputs", "Inputs", "workflow-input"),
            ("secrets", "Secrets", "workflow-secret"),
            ("outputs", "Outputs", "workflow-output"),
        ]

        for key, title, directive in item_lists:
            if item_list := self.call_node.get(key):
                content_node.append(self.format_item_list(title, directive, item_list))
